/* ----------------------------------------------------------------------
 * milo/milo.c -- This program loads the Linux/MIPS kernel and launches it.
 *
 * Copyright (C) 1994 by Waldorf Electronics,
 * written by Ralf Baechle and Andreas Busse
 *
 * Loosly based on bootstrap.c for Linux/68k,
 * Copyright (C) 1993 by Hamish Macdonald, Greg Harp
 *
 * This file is subject to the terms and conditions of the
 * GNU General Public License. See the file COPYING in the
 * main directory of this archive for more details.
 * ---------------------------------------------------------------------- */

#define VERSION   0
#define REVISION 14

#include <stddef.h>
#include <unistd.h>
#include <sys/file.h>
#include <sys/types.h>
#include <asm/io.h>
#include <asm/system.h>
#include <linux/string.h>
#include <linux/tty.h>
#include <linux/a.out.h>
#include <asm/bootinfo.h>
#include <arc/arc.h>

#include "bootstrap.h"
#include "identify.h"

/*
 * Prototypes
 */

void Deskstation_init(void);
void usage(void);

/*
 * Defaults, may be overiden by option or environment.
 */

static char *kernel_name = "multi(0)disk(0)fdisk(0)\\vmlinux";
static char *ramdisk_name = NULL;

extern char *optarg;

extern volatile void launch(char *memptr, u_long kernel_size, u_long rd_size);
struct exec kexec;
char *memptr;

u_long loadaddr    = LOADADDR;	/* mallocinit() needs that */
u_long kernel_base = KERNELBASE; /* whereever that is... */
u_long start_mem   = 0x80000000; /* always true on an ARC system */
u_long mem_size;
u_long rd_size;
u_long kernel_entry;
u_long kernel_bss_end;

struct bootinfo bi;		/* Linux boot info */
struct screen_info si;		/* Linux screen info */
struct DisplayInfo *di;		/* ARC display info */


/* ----------------------------------------------------------------------
 * main()
 * ---------------------------------------------------------------------- */
int main(int argc, char *argv[])
{
  int ch, debugflag = 1, kfd, rfd = -1, i;
  u_long memreq;
  struct nlist *nl;
  u_long kbi_offset, ksi_offset;
  
  /* print the greet message */

  printf("Linux/MIPS ARC Bootstrap Loader V%d.%d\r\n",
	 VERSION,REVISION);
  printf("Copyright (C) Waldorf Electronics, 1994\r\n\r\n");
    
  /* check arguments */

  while ((ch = getopt(argc, argv, "dk:r:")) != EOF)

    switch (ch)
      {
      case 'd':
	debugflag = 1;
	break;
      case 'k':
	kernel_name = optarg;
	break;
      case 'r':
	ramdisk_name = optarg;
	break;
      case '?':
      default:
	usage();
      }

  argc -= optind;
  argv += optind;
    
  /*
   * Identify the system we're running on
   */

  identify();
  
  /*
   * verify that there is enough RAM
   */

  mem_size = bi.memupper - bi.memlower;
  if(mem_size < 0x800000)
    {
      printf("Insufficient Memory to load Linux/MIPS, aborting\n\r");
      exit(5);
    }

  /*
   * open kernel executable and read exec header
   */

  if ((kfd = open (kernel_name, O_RDONLY)) == -1)
    {
      printf ("Unable to open kernel file %s\r\n", kernel_name);
      exit (EXIT_FAILURE);
    }
  if (read (kfd, (void *)&kexec, sizeof(kexec)) != sizeof(kexec))
    {
      printf ("Unable to read exec header from %s\n\r", kernel_name);
      exit (EXIT_FAILURE);
    }
  
  bi.ramdisk_size = 0;		/* default: no ramdisk */
  if (ramdisk_name)
    {
      if ((rfd = open (ramdisk_name, O_RDONLY)) == -1)
	{
	  fprintf (stderr, "Unable to open ramdisk file %s\n\r", ramdisk_name);
	  exit (EXIT_FAILURE);
	}

      /* record ramdisk size  */

      bi.ramdisk_size = (lseek (rfd, 0, L_XTND) + 1023) >> 10;
    }
  
  rd_size = bi.ramdisk_size << 10;
  bi.ramdisk_base = (u_long)start_mem + mem_size - rd_size;

  
  /* find offset to boot_info structure */

  if (!(nl = get_nlist (kernel_name, "_boot_info")))
    {
      perror ("get_nlist");
      exit (EXIT_FAILURE);
    }
  else
    {
      kbi_offset = nl->n_value - kexec.a_entry;
      free (nl);
    }
  
  /* find offset to screen_info structure */

  if (!(nl = get_nlist (kernel_name, "_screen_info")))
    {
      perror ("get_nlist");
      exit (EXIT_FAILURE);
    }
  else
    {
      ksi_offset = nl->n_value - kexec.a_entry;
      free (nl);
    }

  /* Find kernel entry point */

  if (!(nl = get_nlist (kernel_name, "kernel_entry")))
    {
      perror ("get_nlist");
      exit (EXIT_FAILURE);
    }
  else
    {
      kernel_entry = nl->n_value;
      free (nl);
    }

  /* End of bss segment - ramdisk will be placed there */
  
  kernel_bss_end = kernel_base + kexec.a_text + kexec.a_data + kexec.a_bss;
	
  /*
   * Should check whether Kernel & RAMdisk moving doesn't overwrite
   * the bootstraper during startup.
   */

  if (debugflag)
    {
      printf ("\r\n");
      printf ("Kernel text  at 0x%08lx, code size %d\n\r",
	      kernel_base + N_TXTADDR(kexec), kexec.a_text);
      printf ("Kernel data  at 0x%08lx, data size %d\n\r",
	      kernel_base + N_DATADDR(kexec), kexec.a_data );
      printf ("Kernel bss   at 0x%08lx, bss  size %d\n\r",
	      kernel_base + N_BSSADDR(kexec), kexec.a_bss );
      printf ("Kernel entry at 0x%08lx\n\r", kernel_entry );

      if (bi.ramdisk_size)
	{
	  printf ("RAM disk image at    0x%08lx, size is %ldK\n\r",
		  (u_long)memptr + kexec.a_text + kexec.a_data,
		  bi.ramdisk_size);
	  printf ("Ramdisk dest. top is 0x%08lx\n\r", start_mem + mem_size);
	  printf ("Ramdisk low limit is 0x%08lx\n\r",
		  (u_long)(memptr + kexec.a_text + kexec.a_data));
	  printf ("Ramdisk src. top is  0x%08lx\n\r",
		  (u_long)(memptr + kexec.a_text + kexec.a_data)
		  + rd_size);
	}
      printf("\r\n");
      printf ("Hit [Enter] to continue the Linux boot...");
      fflush (stdout);
      getchar();
      printf ("\r\n");
    }

  memreq = kexec.a_text + kexec.a_data + rd_size;
  if (!(memptr = (char *) malloc(memreq)))
    {
      fprintf (stderr, "Unable to allocate %d bytes of memory\n\r",
	       kexec.a_text + kexec.a_data + rd_size);
      exit (EXIT_FAILURE);
    }
  memset(memptr, 0, memreq);
	
  if (lseek (kfd, N_TXTOFF(kexec), L_SET) == -1)
    {
      fprintf (stderr, "Failed to seek to text\n\r");
      free(memptr);
      exit (EXIT_FAILURE);
    }
  if (read (kfd, memptr, kexec.a_text) != kexec.a_text)
    {
      fprintf (stderr, "Failed to read text\n\r");
      free(memptr);
      exit (EXIT_FAILURE);
    }
  if (lseek (kfd, N_DATOFF(kexec), L_SET) == -1)
    {
      fprintf (stderr, "Failed to seek to data\n\r");
      free(memptr);
      exit (EXIT_FAILURE);
    }
  if (read (kfd, memptr + kexec.a_text, kexec.a_data) != kexec.a_data)
    {
      fprintf (stderr, "Failed to read data\n\r");
      free(memptr);
      exit (EXIT_FAILURE);
    }
  close (kfd);
  
  if (rfd != -1)
    {
      if (lseek (rfd, 0, L_SET) == -1)
	{
	  fprintf (stderr, "Failed to seek to beginning of ramdisk file\n\r");
	  free(memptr);
	  exit (EXIT_FAILURE);
	}
      if (read(rfd, memptr+kexec.a_text+kexec.a_data, rd_size) != rd_size)
	{
	  fprintf (stderr, "Failed to read ramdisk file\n\r");
	  free(memptr);
	  exit (EXIT_FAILURE);
	}
      close (rfd);
    }

  /*
   * Initialize the Linux display with ARC data
   */

  if(di = getDisplayInfo())
    {
      si.orig_x            = di->cursor_x;
      si.orig_y            = di->cursor_y + 1;
      si.orig_video_cols   = di->columns;
      si.orig_video_lines  = di->lines;
#if 0
      si.orig_video_mode   = 7;	/* mono */
      si.orig_video_ega_ax = 0; /* unused anyway */
      si.orig_video_ega_bx = 7; /* ega ??? */
      si.orig_video_ega_cx = 0; /* unused anyway */
#endif
    }
  else
    {
      printf("Display Initialization failed. Exiting...\r\n");
      sleep(2);
      exit(EXIT_FAILURE);
    }

  if (bi.machtype == MACH_DESKSTATION_TYNE)
    {
      Deskstation_init();	/* Tynes need some special initialization */
      sleep(2);			/*  wait for things to settle down */
    }
  
  /*
   * copy the boot and screen info structures to the kernel image,
   * then execute the copy-and-go code
   */

  printf("Starting the Kernel...\r\n");
  memcpy ((void *)(memptr + kbi_offset), &bi, sizeof(bi));
  memcpy ((void *)(memptr + ksi_offset), &si, sizeof(si));
  
#if 1
  launch(memptr, kexec.a_text + kexec.a_data, rd_size);
#endif

  /* NOTREACHED */
  
  printf("Kernel returned -- exiting...\r\n");
  sleep(10);
  return(0);
}

/* ----------------------------------------------------------------------
 * Deskstation_init()
 * ---------------------------------------------------------------------- */
void Deskstation_init(void)
{
  /*
   * Grr... The ARC BIOS doesn't setup the hardware like
   * a Intel PC BIOS would do...
   */

  outb_p(0x6,0x3ce);
  outb_p(0xe,0x3cf);

  /* FIXME: This is a hack for Tyne boards. Unfortunally
   * we cannot execute the VGA card bios, so we hardcode
   * some variables here... Will change later.
   */
    
  si.orig_x = 0;
  si.orig_y = 0;
  si.orig_video_page = 52;
  si.orig_video_mode = 3;
  si.orig_video_cols = 80;
  si.orig_video_ega_ax = 4626;
  si.orig_video_ega_bx = 3;
  si.orig_video_ega_cx = 9;
  si.orig_video_lines = 50;
}

/* ----------------------------------------------------------------------
 * usage()
 * ---------------------------------------------------------------------- */
void usage(void)
{
  fprintf (stderr, "Usage:\n"
	   "\tmilo [-d] [-k kernel_executable] [-r ramdisk_file]"
	   " [option...]\n");
  exit (EXIT_FAILURE);
}

/*
 * This is a workaround until I know how to *really* flush all caches..
 * Seems to work now, but I leave that in. Who knows :-)
 */

flushcache()
{
#if 1
  cacheflush();			/* see lib/cachectl.S */
#else  
  FlushAllCaches();		/* ARC BIOS */
#endif  
}
